from functools import partial

import torch
import torch.nn as nn

from timm.models.vision_transformer import PatchEmbed, DropPath, Mlp

from .utils.pos_embed import get_2d_sincos_pos_embed

from omegaconf import OmegaConf
import numpy as np
import math
import scipy.stats as stats
from tqdm import tqdm


def mask_by_random_topk(mask_len, probs, temperature=1.0):
    mask_len = mask_len.squeeze()
    confidence = torch.log(probs) + torch.Tensor(temperature * np.random.gumbel(size=probs.shape)).to(probs.device)
    sorted_confidence, _ = torch.sort(confidence, axis=-1)
    # Obtains cut off threshold given the mask lengths.
    cut_off = sorted_confidence[:, mask_len.long()-1:mask_len.long()]
    # Masks tokens with lower confidence.
    masking = (confidence <= cut_off)
    return masking

class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        # NOTE scale factor was wrong in my original version, can set manually to be compat with prev weights
        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]   # make torchscript happy (cannot use tensor as tuple)

        with torch.amp.autocast('cuda', enabled=False):
            attn = (q.float() @ k.float().transpose(-2, -1)) * self.scale

        attn = attn - torch.max(attn, dim=-1, keepdim=True)[0]
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x, attn


class Block(nn.Module):

    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x, return_attention=False):
        if return_attention:
            _, attn = self.attn(self.norm1(x))
            return attn
        else:
            y, _ = self.attn(self.norm1(x))
            x = x + self.drop_path(y)
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class LabelSmoothingCrossEntropy(nn.Module):
    """ NLL loss with label smoothing.
    """
    def __init__(self, smoothing=0.1):
        super(LabelSmoothingCrossEntropy, self).__init__()
        assert smoothing < 1.0
        self.smoothing = smoothing
        self.confidence = 1. - smoothing

    def forward(self, x: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        logprobs = torch.nn.functional.log_softmax(x, dim=-1)
        nll_loss = -logprobs.gather(dim=-1, index=target.unsqueeze(1))
        nll_loss = nll_loss.squeeze(1)
        smooth_loss = -logprobs.mean(dim=-1)
        loss = self.confidence * nll_loss + self.smoothing * smooth_loss
        return loss


class BertEmbeddings(nn.Module):
    """Construct the embeddings from word, position and token_type embeddings."""

    def __init__(self, vocab_size, hidden_size, max_position_embeddings, dropout=0.1):
        super().__init__()
        self.word_embeddings = nn.Embedding(vocab_size, hidden_size)
        self.position_embeddings = nn.Embedding(max_position_embeddings, hidden_size)

        # self.LayerNorm is not snake-cased to stick with TensorFlow model variable name and be able to load
        # any TensorFlow checkpoint file
        self.LayerNorm = nn.LayerNorm(hidden_size, eps=1e-6)
        self.dropout = nn.Dropout(dropout)
        # position_ids (1, len position emb) is contiguous in memory and exported when serialized
        self.register_buffer("position_ids", torch.arange(max_position_embeddings).expand((1, -1)))

        torch.nn.init.normal_(self.word_embeddings.weight, std=.02)
        torch.nn.init.normal_(self.position_embeddings.weight, std=.02)

    def forward(
        self, input_ids
    ):
        input_shape = input_ids.size()

        seq_length = input_shape[1]

        position_ids = self.position_ids[:, :seq_length]

        inputs_embeds = self.word_embeddings(input_ids)

        position_embeddings = self.position_embeddings(position_ids)
        embeddings = inputs_embeds + position_embeddings

        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class MlmLayer(nn.Module):
    def __init__(self, feat_emb_dim, word_emb_dim, vocab_size):
        super().__init__()
        self.fc = nn.Linear(feat_emb_dim, word_emb_dim)
        self.gelu = nn.GELU()
        self.ln = nn.LayerNorm(word_emb_dim)
        self.bias = nn.Parameter(torch.zeros(1, 1, vocab_size))

    def forward(self, x, word_embeddings):
        mlm_hidden = self.fc(x)
        mlm_hidden = self.gelu(mlm_hidden)
        mlm_hidden = self.ln(mlm_hidden)
        word_embeddings = word_embeddings.transpose(0, 1)
        logits = torch.matmul(mlm_hidden, word_embeddings)
        logits = logits + self.bias
        return logits



class MaskGIT(nn.Module):
    """ Masked Autoencoder with VisionTransformer backbone
    """
    def __init__(self, img_size=256, num_tokens=8192, vae_stride=16, in_channels=3,
                 embed_dim=1024, depth=24, num_heads=16, num_classes=1000,
                 dataset_num=None,
                 mlp_ratio=4., norm_layer=nn.LayerNorm, norm_pix_loss=False,
                 mask_ratio_min=0.05, mask_ratio_max=0.95, mask_ratio_mu=0.5, mask_ratio_std=0.25,
                 label_smoothing=0.1,
                 dataset_label_drop_prob=0.0,
                 label_drop_prob=0.1,  # Classifier-free guidance: probability of dropping class label
                 seq_len=None,
                ):
        super().__init__()

        self.num_classes = num_classes                                             # Number of classes
        if seq_len is None and vae_stride is not None:
            self.seq_len = (img_size // vae_stride) ** 2  
        elif seq_len is not None:
            self.seq_len = seq_len
        else:
            raise ValueError("Either seq_len or vae_stride must be provided")
        self.codebook_size = num_tokens
        vocab_size = self.codebook_size + num_classes + 1 + 1                       # +1 fake class, +1 mask token
        self.fake_class_label = vocab_size - 2
        self.mask_token_label = vocab_size - 1
        self.label_drop_prob = label_drop_prob
        self.dataset_label_drop_prob = dataset_label_drop_prob
        print(f"Dataset label drop probability: {dataset_label_drop_prob}")

        self.token_emb = BertEmbeddings(vocab_size=vocab_size,
                                        hidden_size=embed_dim,
                                        max_position_embeddings=self.seq_len+1,   # seq len + 1 cls token
                                        dropout=0.1)

        # --------------------------------------------------------------------------
        # Dataset ID Embedding (optional)
        self.use_dataset_conditioning = dataset_num is not None
        if self.use_dataset_conditioning:
            self.num_datasets = dataset_num
            self.dataset_emb = nn.Embedding(dataset_num, embed_dim)
            # Fake dataset embedding for unconditional generation
            self.fake_dataset_latent = nn.Parameter(torch.zeros(1, embed_dim))

        # MAGE variant masking ratio
        self.mask_ratio_min = mask_ratio_min
        self.mask_ratio_generator = stats.truncnorm((mask_ratio_min - mask_ratio_mu) / mask_ratio_std,
                                                    (mask_ratio_max - mask_ratio_mu) / mask_ratio_std,
                                                    loc=mask_ratio_mu, scale=mask_ratio_std)

        # --------------------------------------------------------------------------
        # MaskGIT encoder specifics
        dropout_rate = 0.1

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, self.seq_len + 1, embed_dim)) #learnable

        self.blocks = nn.ModuleList([
            Block(embed_dim, num_heads, mlp_ratio, qkv_bias=True, qk_scale=None, norm_layer=norm_layer,
                  drop=dropout_rate, attn_drop=dropout_rate)
            for i in range(depth)])
        self.norm = norm_layer(embed_dim)
        # --------------------------------------------------------------------------

        # --------------------------------------------------------------------------
        # MlmLayer
        self.mlm_layer = MlmLayer(feat_emb_dim=embed_dim, word_emb_dim=embed_dim, vocab_size=vocab_size)

        self.norm_pix_loss = norm_pix_loss

        self.criterion = LabelSmoothingCrossEntropy(smoothing=label_smoothing)

        self.initialize_weights()

    def initialize_weights(self):
        # initialization

        # timm's trunc_normal_(std=.02) is effectively normal_(std=0.02) as cutoff is too big (2.)
        torch.nn.init.normal_(self.cls_token, std=.02)
        torch.nn.init.normal_(self.pos_embed, std=.02)

        # initialize dataset embedding if used
        if self.use_dataset_conditioning:
            torch.nn.init.normal_(self.dataset_emb.weight, std=.02)
            torch.nn.init.normal_(self.fake_dataset_latent, std=.02)

        # initialize nn.Linear and nn.LayerNorm
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            # we use xavier_uniform following official JAX ViT:
            torch.nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward_encoder(self, x, y=None, dataset_id=None):
        """
        Args:
            x: (bsz, seq_len) token indices
            y: (bsz,) class labels, optional. If None, uses fake_class_label
            dataset_id: (bsz,) dataset IDs, optional. Only used if dataset_num was provided in __init__
        """
        token_indices = x
        gt_indices = token_indices.clone().detach().long()

        # masking
        bsz, seq_len = token_indices.size()
        mask_ratio_min = self.mask_ratio_min
        mask_rate = self.mask_ratio_generator.rvs(1)[0]

        num_dropped_tokens = int(np.ceil(seq_len * mask_ratio_min))
        num_masked_tokens = int(np.ceil(seq_len * mask_rate))

        # it is possible that two elements of the noise is the same, so do a while loop to avoid it
        while True:
            noise = torch.rand(bsz, seq_len, device=x.device)  # noise in [0, 1]
            sorted_noise, _ = torch.sort(noise, dim=1)  # ascend: small is remove, large is keep
            cutoff_drop = sorted_noise[:, num_dropped_tokens-1:num_dropped_tokens]
            cutoff_mask = sorted_noise[:, num_masked_tokens-1:num_masked_tokens]
            token_drop_mask = (noise <= cutoff_drop).float()
            token_all_mask = (noise <= cutoff_mask).float()
            if token_drop_mask.sum() == bsz*num_dropped_tokens and token_all_mask.sum() == bsz*num_masked_tokens:
                break
            else:
                print("Rerandom the noise!")
        # print(mask_rate, num_dropped_tokens, num_masked_tokens, token_drop_mask.sum(dim=1), token_all_mask.sum(dim=1))
        token_indices[token_all_mask.nonzero(as_tuple=True)] = self.mask_token_label
        # print("Masekd num token:", torch.sum(token_indices == self.mask_token_label, dim=1))

        # Determine class token labels
        # Class labels are in range [codebook_size, codebook_size + num_classes)
        if y is not None:
            # Apply label dropout for classifier-free guidance during training
            if self.training and self.label_drop_prob > 0:
                drop_mask = torch.rand(bsz, device=x.device) < self.label_drop_prob
                class_labels = torch.where(
                    drop_mask,
                    torch.full_like(y, self.fake_class_label),
                    self.codebook_size + y  # Map class label to vocabulary index
                )
            else:
                class_labels = self.codebook_size + y
        else:
            class_labels = torch.full((bsz,), self.fake_class_label, device=x.device, dtype=torch.long)

        # concate class token
        token_indices = torch.cat([torch.zeros(token_indices.size(0), 1).to(token_indices.device), token_indices], dim=1)
        token_indices[:, 0] = class_labels
        token_drop_mask = torch.cat([torch.zeros(token_indices.size(0), 1).to(token_indices.device), token_drop_mask], dim=1)
        token_all_mask = torch.cat([torch.zeros(token_indices.size(0), 1).to(token_indices.device), token_all_mask], dim=1)
        token_indices = token_indices.long()
        # bert embedding
        input_embeddings = self.token_emb(token_indices)
        # print("Input embedding shape:", input_embeddings.shape)
        bsz, seq_len, emb_dim = input_embeddings.shape

        # Add dataset embedding to class token (position 0) if dataset conditioning is enabled
        if self.use_dataset_conditioning:
            zeros_ds = torch.zeros(
                input_embeddings.size(0), 1, input_embeddings.size(-1),
                device=input_embeddings.device, dtype=input_embeddings.dtype
            )
            input_embeddings = torch.cat([zeros_ds, input_embeddings], dim=1)
            drop_prob = self.dataset_label_drop_prob
            if dataset_id is not None:
                if drop_prob > 0:
                    drop_mask = torch.rand(bsz, device=x.device) < drop_prob
                    dataset_embedding = torch.where(
                        drop_mask.unsqueeze(-1),
                        self.fake_dataset_latent.expand(bsz, -1),
                        self.dataset_emb(dataset_id)
                    )
                else:
                    dataset_embedding = self.dataset_emb(dataset_id)
            else:
                dataset_embedding = self.fake_dataset_latent.expand(bsz, -1)

            # Add dataset embedding
            input_embeddings[:, 0] = dataset_embedding
        print(f"input_embeddings.shape: {input_embeddings.shape}")
        # No dropping needed because we use BEiT style architecture
        # dropping
        # token_keep_mask = 1 - token_drop_mask
        # input_embeddings_after_drop = input_embeddings[token_keep_mask.nonzero(as_tuple=True)].reshape(bsz, -1, emb_dim)
        # print("Input embedding after drop shape:", input_embeddings_after_drop.shape)

        # apply Transformer blocks
        x = input_embeddings
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)
        # print("Encoder representation shape:", x.shape)

        return x, gt_indices, token_drop_mask, token_all_mask

    def forward_loss(self, gt_indices, logits, mask):
        bsz, seq_len = gt_indices.size()
        # logits and mask are with seq_len+1 but gt_indices is with seq_len
        if self.use_dataset_conditioning:
            logits = logits[:, 2:, :self.codebook_size].reshape(bsz*seq_len, -1)
            gt_indices = gt_indices.reshape(bsz*seq_len)
        else:
            logits = logits[:, 1:, :self.codebook_size].reshape(bsz*seq_len, -1)
            gt_indices = gt_indices.reshape(bsz*seq_len)


        loss = self.criterion(logits, gt_indices)
        loss = loss.reshape(bsz, seq_len)
        loss = (loss * mask[:, 1:]).sum() / mask[:, 1:].sum()  # mean loss on removed patches
        return loss

    def forward(self, imgs, y=None, dataset_id=None):
        """
        Args:
            imgs: (bsz, seq_len) token indices
            y: (bsz,) class labels, optional. If provided, uses class conditioning.
            dataset_id: (bsz,) dataset IDs, optional. Only used if dataset_num was provided in __init__
        """
        latent, gt_indices, token_drop_mask, token_all_mask = self.forward_encoder(imgs, y=y, dataset_id=dataset_id)
        word_embeddings = self.token_emb.word_embeddings.weight.data.detach()
        logits = self.mlm_layer(latent, word_embeddings)
        loss = self.forward_loss(gt_indices, logits, token_all_mask)
        return loss


    def sample(self, bsz, device, input_token_indices=None, mask=None, y=None, dataset_id=None, num_iter=12, choice_temperature=4.5, verbose=False, cfg=1.0):
        """
        Sample tokens using iterative refinement (MaskGIT sampling).
        
        Supports two modes:
        1. Inpainting: Provide input_token_indices and mask
        2. Generation from scratch: Provide bsz and seq_len (or they will be inferred)
        
        Args:
            input_token_indices: (bsz, seq_len) LongTensor of input tokens, optional.
                If None, generation from scratch mode is used.
            mask: (bsz, seq_len) ByteTensor or BoolTensor where 1 = to inpaint (masked), 0 = keep as is.
                If None and input_token_indices is None, all tokens are masked (generation from scratch).
                If None but input_token_indices is provided, assumes all tokens are masked.
            y: (bsz,) class labels, optional. If None, uses fake_class_label (unconditional)
            dataset_id: (bsz,) dataset IDs, optional. Only used if dataset_num was provided in __init__
            num_iter: number of iterative refinement steps
            choice_temperature: temperature for token selection
            verbose: whether to show progress bar
            cfg: classifier-free guidance scale. If > 1.0, performs CFG by doubling batch size
            bsz: batch size, required if input_token_indices is None
            seq_len: sequence length, required if input_token_indices is None. If None, uses self.seq_len
        """        
        # Determine mode and initialize accordingly
        if input_token_indices is None:
            # Generation from scratch mode
            # Initialize all tokens as masked
            token_indices = torch.full((bsz, self.seq_len), self.mask_token_label, device=device, dtype=torch.long)
            mask = torch.ones((bsz, self.seq_len), device=device, dtype=torch.bool)
        else:
            # Inpainting mode
            bsz, seq_len = input_token_indices.shape
            input_token_indices = input_token_indices.to(device)
            
            if mask is None:
                # If mask not provided, assume all tokens are masked (generation from scratch with given shape)
                mask = torch.ones((bsz, seq_len), device=device, dtype=torch.bool)
            else:
                mask = mask.bool().to(device)
            
            # Initialize token indices: masked positions get mask_token_label, others get given input
            token_indices = torch.where(
                mask,
                torch.full_like(input_token_indices, self.mask_token_label),
                input_token_indices
            )
        bsz, seq_len = token_indices.shape
        use_cfg = cfg > 1.0 and y is not None
        original_bsz = bsz  # Keep track of original batch size
        
        # Prepare class labels
        if y is not None:
            class_labels = (self.codebook_size + y).to(device)  # Map class to vocabulary index
        else:
            class_labels = torch.full((bsz,), self.fake_class_label, device=device, dtype=torch.long)
        
        # Prepare dataset embedding (optional)
        dataset_embedding = None
        if self.use_dataset_conditioning:
            drop_prob = self.dataset_label_drop_prob
            if dataset_id is not None:
                dataset_embedding = self.dataset_emb(dataset_id.to(device))
                if drop_prob > 0:
                    drop_mask = torch.rand(bsz, device=device) < drop_prob
                    dataset_embedding = torch.where(
                        drop_mask.unsqueeze(-1),
                        self.fake_dataset_latent.expand(bsz, -1).to(device),
                        dataset_embedding
                    )
            else:
                dataset_embedding = self.fake_dataset_latent.expand(bsz, -1).to(device)
        
        # Classifier-free guidance: duplicate batch ONCE before the loop
        if use_cfg:
            token_indices = torch.cat([token_indices, token_indices], dim=0)
            mask = torch.cat([mask, mask], dim=0)
            class_labels = torch.cat([
                class_labels,
                torch.full((bsz,), self.fake_class_label, device=device, dtype=torch.long)
            ], dim=0)
            if dataset_embedding is not None:
                dataset_embedding = torch.cat([
                    dataset_embedding,
                    self.fake_dataset_latent.expand(bsz, -1).to(device)
                ], dim=0)
            bsz = bsz * 2

        _CONFIDENCE_OF_KNOWN_TOKENS = float("inf")

        for step in tqdm(range(num_iter), disable=not verbose):
            # 2. Prepend class token
            token_indices_with_cls = torch.cat(
                [class_labels.unsqueeze(1), token_indices], dim=1
            )

            # 3. Token embedding → transformer → norm
            input_embeddings = self.token_emb(token_indices_with_cls)
            
            # Add dataset embedding to class token (position 0) if dataset conditioning is enabled
            if dataset_embedding is not None:
                zeros_ds = torch.zeros(
                    input_embeddings.size(0), 1, input_embeddings.size(-1),
                    device=input_embeddings.device, dtype=input_embeddings.dtype
                )
                input_embeddings = torch.cat([zeros_ds, input_embeddings], dim=1)
                input_embeddings[:, 0] = input_embeddings[:, 0] + dataset_embedding
            
            x = input_embeddings
            for blk in self.blocks:
                x = blk(x)
            latent = self.norm(x)

            # 4. Remove class token from output
            if self.use_dataset_conditioning:
                latent = latent[:, 2:, :]  # [bsz, seq_len, hidden_dim]
            else:
                latent = latent[:, 1:, :]  # [bsz, seq_len, hidden_dim]

            # 5. Get logits from MLM head
            word_embeddings = self.token_emb.word_embeddings.weight.detach()
            logits = self.mlm_layer(latent, word_embeddings)
            logits = logits[:, :, :self.codebook_size]  # [bsz, seq_len, vocab_size]

            # 6. Apply classifier-free guidance if enabled
            if use_cfg:
                cond_logits, uncond_logits = logits.chunk(2, dim=0)
                # CFG: logits = uncond_logits + cfg * (cond_logits - uncond_logits)
                logits = uncond_logits + cfg * (cond_logits - uncond_logits)
                # After CFG, logits have original batch size, so reduce mask, token_indices, and dataset_embedding to match
                mask = mask[:original_bsz]
                token_indices = token_indices[:original_bsz]
                if dataset_embedding is not None:
                    dataset_embedding = dataset_embedding[:original_bsz]
                bsz = original_bsz

            # 7. Sample tokens
            sample_dist = torch.distributions.Categorical(logits=logits)
            sampled_ids = sample_dist.sample()  # [bsz, seq_len] (original batch size after CFG)

            # 8. Keep known tokens fixed (using original batch size mask and token_indices)
            sampled_ids = torch.where(mask, sampled_ids, token_indices)

            # 9. Compute mask ratio for next iteration (cosine schedule)
            ratio = (step + 1) / num_iter
            mask_ratio = np.cos(np.pi / 2. * ratio)
            mask_ratio = torch.tensor(mask_ratio, device=device)

            # 10. Confidence scores for predicted tokens
            probs = torch.nn.functional.softmax(logits, dim=-1)
            selected_probs = torch.gather(probs, dim=-1, index=sampled_ids.unsqueeze(-1)).squeeze(-1)

            # 11. Force known tokens to have high confidence so they're never remasked
            selected_probs = torch.where(mask, selected_probs, torch.full_like(selected_probs, _CONFIDENCE_OF_KNOWN_TOKENS))

            # 12. Determine how many tokens to re-mask
            unknown_counts = mask.sum(dim=1, keepdim=True)  # how many masked tokens per example
            mask_len = torch.floor(self.seq_len * mask_ratio).to(device)
            mask_len = torch.clamp(mask_len, min=1)
            mask_len = torch.min(mask_len, unknown_counts - 1)
            mask_len = torch.clamp(mask_len, min=1)

            # 13. Create new mask for next iteration (low-confidence tokens get re-masked)
            masking = mask_by_random_topk(mask_len[0], selected_probs, choice_temperature * (1 - ratio))
            # Combine with original mask so we don't touch known tokens
            new_mask = torch.where(mask, masking, torch.zeros_like(masking, dtype=torch.bool))

            # 14. Update token indices: re-mask low-confidence tokens
            token_indices = torch.where(new_mask, self.mask_token_label, sampled_ids)

            mask = new_mask  # update for next round
            
            # 15. If CFG was used, duplicate for next iteration (after all computations with original batch size)
            if use_cfg:
                token_indices = torch.cat([token_indices, token_indices], dim=0)
                mask = torch.cat([mask, mask], dim=0)
                if dataset_embedding is not None:
                    dataset_embedding = torch.cat([
                        dataset_embedding[:original_bsz],
                        self.fake_dataset_latent.expand(original_bsz, -1).to(device)
                    ], dim=0)
                bsz = bsz * 2
        
        # If CFG was used, return only the first half (conditional samples)
        if use_cfg:
            sampled_ids = sampled_ids[:original_bsz]
        
        return sampled_ids






import torch 
import torch.nn as nn
from functools import partial

from medlat.registry import register_model
from .mar import MAR

@register_model("mar.b")
def MAR_B(**kwargs):
    model = MAR(
        encoder_embed_dim=768, encoder_depth=12, encoder_num_heads=12,
        decoder_embed_dim=768, decoder_depth=12, decoder_num_heads=12,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


@register_model("mar.l")
def MAR_L(**kwargs):
    model = MAR(
        encoder_embed_dim=1024, encoder_depth=16, encoder_num_heads=16,
        decoder_embed_dim=1024, decoder_depth=16, decoder_num_heads=16,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


@register_model("mar.h")
def MAR_H(**kwargs):
    model = MAR(
        encoder_embed_dim=1280, encoder_depth=20, encoder_num_heads=16,
        decoder_embed_dim=1280, decoder_depth=20, decoder_num_heads=16,
        mlp_ratio=4, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
    return model


import torch 
import torch.nn as nn
from typing import Tuple
from .modules import ImagingMaskedEncoder, ImagingMaskedDecoder, ReconstructionCriterion
from .utils.imaging_model_related import Masker, sincos_pos_embed, patchify, unpatchify
from medlat.utils import init_from_ckpt
from medlat.registry import register_model



@register_model("token.vita.reconmae")
class ReconMAE(nn.Module):
    def __init__(self, 
                img_shape: tuple,
                patch_embed_cls: str = "PatchEmbed",
                patch_size: Tuple[int, ...] = (5, 8, 8),
                patch_in_channels: int = 1,
                pixel_unshuffle_scale: int = 1,
                mask_type: str = "random",
                mask_ratio: float = 0.0,
                circular_pe: bool = False,
                use_enc_pe: bool = True,
                mask_loss: bool = True,
                shift_size: Tuple[int, ...] = (0, 0, 0),
                enc_embed_dim: int = 1025,
                enc_depth: int = 6,
                enc_num_heads: int = 5,
                mlp_ratio: float = 4.0,
                grad_checkpointing: bool = False,
                use_both_axes: bool = False,
                dec_embed_dim: int = 1025,
                dec_num_heads: int = 5,
                dec_depth: int = 2,
                ckpt_path: str = None,
                loss_types: Tuple[str, ...] = ("mse"),
                loss_weights: Tuple[float, ...] = (1.0),
                *args, **kwargs):
        super().__init__(*args, **kwargs)

        # self.data_view = self.hparams.val_dset.view
        # use_both_axes = True if self.hparams.val_dset.get_view() == 2 else False # For positional embedding
        # img_shape = self.hparams.val_dset[0][0].shape
        print(f"img_shape: {img_shape}, use_both_axes: {use_both_axes}, mask_ratio: {mask_ratio}")
        self.patch_size = patch_size
        self.img_shape = img_shape

        self.encoder_imaging = ImagingMaskedEncoder(img_shape=img_shape, 
                                                    use_both_axes=use_both_axes,
                                                    patch_embed_cls=patch_embed_cls,
                                                    patch_size=patch_size,
                                                    patch_in_channels=patch_in_channels,
                                                    pixel_unshuffle_scale=pixel_unshuffle_scale,
                                                    mask_type=mask_type,
                                                    mask_ratio=mask_ratio,
                                                    circular_pe=circular_pe,
                                                    use_enc_pe=use_enc_pe,
                                                    mask_loss=mask_loss,
                                                    shift_size=shift_size,
                                                    enc_embed_dim=enc_embed_dim,
                                                    enc_depth=enc_depth,
                                                    enc_num_heads=enc_num_heads,
                                                    mlp_ratio=mlp_ratio,
                                                    grad_checkpointing=grad_checkpointing,
                                                    )
        self.decoder_imaging = ImagingMaskedDecoder(decoder_num_patches=self.encoder_imaging.patch_embed.num_patches, 
                                                    grid_size=self.encoder_imaging.patch_embed.grid_size,                               
                                                    head_out_dim=self.encoder_imaging.patch_p_num,
                                                    dec_num_heads=dec_num_heads,
                                                    dec_depth=dec_depth,
                                                    mlp_ratio=mlp_ratio,
                                                    use_enc_pe=use_enc_pe,
                                                    enc_embed_dim=enc_embed_dim,
                                                    dec_embed_dim=dec_embed_dim,
                                                    use_both_axes=use_both_axes,
                                                    )
        self.reconstruction_criterion = ReconstructionCriterion(loss_types=loss_types, loss_weights=loss_weights)

        if ckpt_path is not None:
            init_from_ckpt(self, ckpt_path)
        
    def forward(self, x):
        x, mask, ids_restore = self.encoder_imaging(x)
        x = self.decoder_imaging(x, ids_restore)
        b = x.shape[0]
        img_shape = (b, *self.img_shape)
        x = unpatchify(x, im_shape=img_shape, patch_size=self.patch_size)
        return x, mask
        
    def training_step(self, batch, batch_idx, mode="train"):
        imgs, sub_idx = batch

        pred_patches, mask = self.forward(imgs)

        imgs_patches = patchify(imgs, patch_size=self.hparams.patch_size)
        loss_dict = self.reconstruction_criterion(pred_patches, imgs_patches, mask) 
        psnr_value = calculate_psnr(pred_patches, imgs_patches, mask, replace_with_gt=True)
        
        # Logging metrics and median
        self.log_recon_metrics(loss_dict, psnr_value, mode=mode)

        if mode == "val": # For checkpoint tracking
            self.module_logger.update_metric_item(f"{mode}_PSNR", psnr_value, mode=mode) 
            
        log_rate = eval(f"self.hparams.{mode}_log_rate")
        if self.current_epoch > 0 and ((self.current_epoch + 1) % log_rate == 0):
            if (sub_idx == 0).any():
                i = (sub_idx == 0).argwhere().squeeze().item()
                self.log_recon_videos(pred_patches[i], imgs[i], sub_idx[i], mode=mode)
            if (sub_idx == 1).any():
                i = (sub_idx == 1).argwhere().squeeze().item()
                self.log_recon_videos(pred_patches[i], imgs[i], sub_idx[i], mode=mode)
        return loss_dict["loss"]
    
    @torch.no_grad()
    def validation_step(self, batch, batch_idx):
        _ = self.training_step(batch, batch_idx, mode="val")

    @torch.no_grad()
    def test_step(self, batch, batch_idx):
        imgs, sub_idx = batch

        pred_patches, mask = self.forward(imgs)

        pred_imgs = unpatchify(pred_patches, im_shape=imgs.shape, patch_size=self.hparams.patch_size)
        psnr_value = calculate_psnr(pred_imgs, imgs, reduction="none") # (B, S)
        psnr_value = np.mean(psnr_value, axis=2)
        self.test_psnr.append(psnr_value)

        # Save sample images
        if (sub_idx == 0).any():
            i = (sub_idx == 0).argwhere().squeeze().item()
            sample_imgs = imgs[i]
            sample_recon_imgs = pred_imgs[i]
            self.log_recon_videos(pred_patches[i], imgs[i], sub_idx[i], mode="test")

            if self.hparams.test_sample_path is not None:
                Path(self.hparams.test_sample_path).mkdir(parents=True, exist_ok=True)
                view_name = self.hparams.test_sample_path.split('_')[-1]
                sample_imgs = sample_imgs.detach().cpu().numpy()
                sample_recon_imgs = sample_recon_imgs.detach().cpu().numpy()
                for k in range(sample_imgs.shape[0]):
                    gt = sample_imgs[k, 13]
                    recon = sample_recon_imgs[k, 13]
                    plt.imsave(Path(self.hparams.test_sample_path) / f"gt_0_s{k}_t13.png", gt, cmap="gray")
                    plt.imsave(Path(self.hparams.test_sample_path) / f"recon_{view_name}_0_s{k}_t13.png", recon, cmap="gray")

    def on_test_epoch_end(self) -> None:
        test_psnr = np.concatenate(self.test_psnr, axis=0) # (#test, S)
        psnr_mean = np.mean(test_psnr)
        psnr_std = np.std(test_psnr)
        # Short-axis
        if self.data_view == 0:
            psnr_sax_mean = psnr_mean
            psnr_sax_std = psnr_std
            psnr_lax_mean = 0
            psnr_lax_std = 0
        # Long-axis
        elif self.data_view == 1:
            psnr_sax_mean = 0
            psnr_sax_std = 0
            psnr_lax_mean = psnr_mean
            psnr_lax_std = psnr_std
        elif self.data_view == 2:
            psnr_lax_mean = np.mean(test_psnr[:, :3])
            psnr_lax_std = np.std(test_psnr[:, :3])
            psnr_sax_mean = np.mean(test_psnr[:, 3:])
            psnr_sax_std = np.std(test_psnr[:, 3:])
        results = {"psnr_mean": psnr_mean, "psnr_std": psnr_std, 
                   "psnr_sax_mean": psnr_sax_mean, "psnr_sax_std": psnr_sax_std, 
                   "psnr_lax_mean": psnr_lax_mean, "psnr_lax_std": psnr_lax_std, }
        
        # Table logging
        columns = list(results.keys())
        test_table = wandb.Table(columns=columns)
        test_table.add_data(*[results[col] for col in columns])
        wandb.log({"Evaluation_table": test_table})

        if self.hparams.test_psnr_path is not None:
            Path(self.hparams.test_psnr_path).parent.mkdir(parents=True, exist_ok=True)
            with open(self.hparams.test_psnr_path, "wb") as file:
                pickle.dump(results, file)

        self.wandb_log(self.current_epoch, mode="test")
                
    @torch.no_grad()
    def generate_latents(self, data_loader, 
                         token_path: str, 
                         tsne_map_path: str,
                         save_all_patch_tokens: bool = True,
                         save_tsne: bool = True,
                         ):
        latents = []
        for i, batch in tqdm(enumerate(data_loader), total=len(data_loader), desc="Generating latent codes"):
            imgs, sub_idx = batch
            if torch.cuda.is_available():
                imgs = imgs.to("cuda")
            sub_path = data_loader.dataset.subject_paths[sub_idx]
            sub_id = sub_path.parent.name
            enc_output_latent, _, ids_restore = self.encoder_imaging(imgs)
            cls_t = enc_output_latent[:, 0, :]
            all_t_ = enc_output_latent[:, 1:, :]
            
            # Restore the order of the tokens
            all_t_restore = torch.gather(all_t_, dim=1, index=ids_restore.unsqueeze(-1).repeat(1, 1, all_t_.shape[-1]))
            
            # Take the average of tokens across spatial dimensions
            B, S, T = imgs.shape[:3]
            num_t_patches = T // self.hparams.patch_size[0]
            all_t_restore = all_t_restore.reshape(B, S, num_t_patches, -1, all_t_restore.shape[-1])
            all_t_restore = all_t_restore.moveaxis(1, 2)
            all_t = all_t_restore.reshape(B, num_t_patches, -1, all_t_restore.shape[-1]).mean(2)
            
            if save_all_patch_tokens:
                latents.append({'subj_id': np.array(int(sub_id)).reshape(1, 1), 
                                'cls_token': cls_t.detach().cpu().numpy(), 
                                'all_token': all_t.detach().cpu().numpy()})
            else:
                latents.append({'subj_id': np.array(int(sub_id)).reshape(1, 1), 
                                'cls_token': cls_t.detach().cpu().numpy(),})
        
        # Save the latent codes
        if not save_all_patch_tokens:
            cls_token = np.concatenate([i['cls_token'] for i in latents])
            subj_id = np.concatenate([i['subj_id'] for i in latents]).reshape(-1)
            np.savez(token_path, cls_token=cls_token, subj_id=subj_id)
        else:
            cls_token = np.concatenate([i['cls_token'] for i in latents])
            subj_id = np.concatenate([i['subj_id'] for i in latents]).reshape(-1)
            all_token = np.concatenate([i['all_token'] for i in latents])
            np.savez(token_path, cls_token=cls_token, all_token=all_token, subj_id=subj_id)
        
        if save_tsne:
        # Save the t-SNE embeddings
            scaled_avg = StandardScaler().fit_transform(np.mean(all_token, axis=1))
            scaled_cls = StandardScaler().fit_transform(cls_token)
            scaled_tmp = StandardScaler().fit_transform(all_token.reshape(-1, all_token.shape[2]))
            
            tsne_map_cls = TSNE(n_components=3, perplexity=5, learning_rate="auto").fit_transform(scaled_cls)
            tsne_map_avg = TSNE(n_components=3, perplexity=5, learning_rate="auto").fit_transform(scaled_avg)
            tsne_map_tmp = TSNE(n_components=3, perplexity=5, learning_rate="auto").fit_transform(scaled_tmp)
                
            np.savez(tsne_map_path, 
                    tsne_map_cls=tsne_map_cls, 
                    tsne_map_avg=tsne_map_avg, 
                    tsne_map_tmp=tsne_map_tmp, 
                    subj_id=subj_id)



from medlat.first_stage.continuous.modules.ldm_modules import Encoder as LDMEncoder, Decoder as LDMDecoder
from medlat.first_stage.continuous.modules.maisi_modules import MaisiEncoder, MaisiDecoder
from medlat.first_stage.continuous.modules.dcae_modules.dcae_modules import DCAEEncoder, DCAEDecoder
from medlat.registry import register_model
from medlat.first_stage.continuous.vae_models import AutoencoderKL, AutoencoderKLTransformer
from medlat.modules.alignments import *
from medlat.modules.vit_core import GenericViTEncoder, GenericViTDecoder


@register_model(f"continuous.aekl.f4_d3", 
code_url="https://github.com/CompVis/latent-diffusion/blob/main/models/first_stage_models/kl-f4/config.yaml", 
paper_url="https://arxiv.org/pdf/2112.10752",)
def AEKL_f4_d3(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=3,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 2, 4],
    num_res_blocks=2,
    attn_resolutions=[],
    dropout=0.0,
    **kwargs):
    """
    AutoencoderKL with compression factor 4
    Args:
        dims (int): Number of dimensions (2 for 2D, 3 for 3D)
    """
    encoder = LDMEncoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    decoder = LDMDecoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    return AutoencoderKL(encoder=encoder, decoder=decoder, **kwargs)

@register_model(f"continuous.aekl.f4_d8")
def AEKL_f4_d8(**kwargs):
    return AEKL_f4_d3(z_channels=8, **kwargs)

@register_model(f"continuous.aekl.f4_d16")
def AEKL_f4_d16(**kwargs):
    return AEKL_f4_d3(z_channels=16, **kwargs)

@register_model(f"continuous.aekl.f4_d32")
def AEKL_f4_d32(**kwargs):
    return AEKL_f4_d3(z_channels=32, **kwargs)

@register_model(f"continuous.aekl.f8_d4", 
code_url="https://github.com/CompVis/latent-diffusion/blob/main/models/first_stage_models/kl-f8/config.yaml", 
paper_url="https://arxiv.org/pdf/2112.10752",)
def AEKL_f8_d4(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=4,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 2, 4, 4],
    num_res_blocks=2,
    attn_resolutions=[],
    dropout=0.0,
    **kwargs):
    """
    AutoencoderKL with compression factor 8.
    Args:
        dims (int): Number of dimensions (2 for 2D, 3 for 3D)
        kwargs: Can include 'ddconfig' dict or any ddconfig key directly
    """
    encoder = LDMEncoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    decoder = LDMDecoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    return AutoencoderKL(encoder=encoder, decoder=decoder, **kwargs)


@register_model(f"continuous.aekl.f8_d8")
def AEKL_f8_d8(**kwargs):
    return AEKL_f8_d4(z_channels=8, **kwargs)

@register_model(f"continuous.aekl.f8_d16")
def AEKL_f8_d16(**kwargs):
    return AEKL_f8_d4(z_channels=16, **kwargs)

@register_model(f"continuous.aekl.f8_d32")
def AEKL_f8_d32(**kwargs):
    return AEKL_f8_d4(z_channels=32, **kwargs)

@register_model(f"continuous.aekl.f16_d8", 
code_url="https://github.com/CompVis/latent-diffusion/blob/main/configs/autoencoder/autoencoder_kl_16x16x16.yaml", 
paper_url="https://arxiv.org/pdf/2112.10752",
description="ATTENTION: There are two different official configurations with z=8 and z=16 depending on the repo, we use z=8 here.")
def AEKL_f16_d8(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=8,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 1, 2, 2, 4],
    num_res_blocks=2,
    attn_resolutions=[16],
    dropout=0.0,
    **kwargs):
    """
    AutoencoderKL with compression factor 16
    Args:
        dims (int): Number of dimensions (2 for 2D, 3 for 3D)
    """
    encoder = LDMEncoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    decoder = LDMDecoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    return AutoencoderKL(encoder=encoder, decoder=decoder, **kwargs)

@register_model(f"continuous.aekl.f16_d16")
def AEKL_f16_d16(**kwargs):
    return AEKL_f16_d8(z_channels=16, **kwargs)

@register_model(f"continuous.aekl.f16_d32")
def AEKL_f16_d32(**kwargs):
    return AEKL_f16_d8(z_channels=32, **kwargs)

@register_model(f"continuous.aekl.f16_d64")
def AEKL_f16_d64(**kwargs):
    return AEKL_f16_d8(z_channels=64, **kwargs)

@register_model(f"continuous.aekl.f32_d64", 
code_url="https://github.com/CompVis/latent-diffusion/blob/main/models/first_stage_models/kl-f32/config.yaml",
paper_url="https://arxiv.org/pdf/2112.10752",)
def AEKL_f32_d64(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=64,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 1, 2, 2, 4, 4],
    num_res_blocks=2,
    attn_resolutions=[16, 8],
    dropout=0.0,
    **kwargs):
    """
    AutoencoderKL with compression factor 32
    Args:
        dims (int): Number of dimensions (2 for 2D, 3 for 3D)
    """
    encoder = LDMEncoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    decoder = LDMDecoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    return AutoencoderKL(encoder=encoder, decoder=decoder, **kwargs)

@register_model(f"continuous.maisi.f4_d3", 
code_url="https://github.com/monai/monai/blob/main/monai/networks/nets/maisi_autoencoderkl.py",
paper_url="https://arxiv.org/pdf/2507.05622",)
def Maisi_f4_d4(
    z_channels: int = 4,
    spatial_dims: int = 3,
    in_channels: int = 1,
    out_ch: int = 1,
    num_res_blocks: list[int] = [2, 2, 2],
    num_channels: list[int] = [64, 128, 256],
    attention_levels: list[bool] = [False, False, False],
    norm_num_groups: int = 32,
    norm_eps: float = 1e-6,
    with_encoder_nonlocal_attn: bool = False,
    with_decoder_nonlocal_attn: bool = False,
    include_fc: bool = False,
    use_combined_linear: bool = False,
    use_flash_attention: bool = False,
    use_convtranspose: bool = False,
    num_splits: int = 1,
    dim_split: int = 1,
    double_z: bool = True,
    norm_float16: bool = True,
    print_info: bool = False,
    save_mem: bool = True,
    **kwargs):

    encoder = MaisiEncoder(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            num_channels=num_channels,
            out_channels=z_channels,
            num_res_blocks=num_res_blocks,
            norm_num_groups=norm_num_groups,
            norm_eps=norm_eps,
            attention_levels=attention_levels,
            with_nonlocal_attn=with_encoder_nonlocal_attn,
            include_fc=include_fc,
            use_combined_linear=use_combined_linear,
            use_flash_attention=use_flash_attention,
            num_splits=num_splits,
            dim_split=dim_split,
            norm_float16=norm_float16,
            double_z=double_z,
            print_info=print_info,
            save_mem=save_mem,
        )

    decoder = MaisiDecoder(
            spatial_dims=spatial_dims,
            num_channels=num_channels,
            in_channels=z_channels,
            out_channels=out_ch,
            num_res_blocks=num_res_blocks,
            norm_num_groups=norm_num_groups,
            norm_eps=norm_eps,
            attention_levels=attention_levels,
            with_nonlocal_attn=with_decoder_nonlocal_attn,
            include_fc=include_fc,
            use_combined_linear=use_combined_linear,
            use_flash_attention=use_flash_attention,
            use_convtranspose=use_convtranspose,
            num_splits=num_splits,
            dim_split=dim_split,
            norm_float16=norm_float16,
            double_z=double_z,
            print_info=print_info,
            save_mem=save_mem,
        )

    return AutoencoderKL(encoder=encoder, decoder=decoder, **kwargs)

@register_model(f"continuous.medvae.f8_d16")
def MedVAE_f8_d16(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=16,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 2, 4, 4],
    num_res_blocks=2,
    attn_resolutions=[],
    dropout=0.0,
    **kwargs
):

    encoder = LDMEncoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    decoder = LDMDecoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)

    alignment = VFFoundationAlignment(latent_channels=z_channels, foundation_type="biomedclip")
    return AutoencoderKL(encoder=encoder, decoder=decoder, alignment=alignment, **kwargs)


@register_model(f"continuous.medvae.f8_d32")
def MedVAE_f8_d32(**kwargs):
    return MedVAE_f8_d16(z_channels=32, **kwargs)


@register_model(f"continuous.vavae.f8_d32_dinov2")
def VAVAE_f8_d32_dinov2(**kwargs):
    return VAVAE_f8_d16_dinov2(z_channels=32, **kwargs)

@register_model(f"continuous.vavae.f8_d16_dinov2")
def VAVAE_f8_d16_dinov2(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=16,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 2, 4, 4],
    num_res_blocks=2,
    attn_resolutions=[],
    dropout=0.0,
    **kwargs
):

    encoder = LDMEncoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    decoder = LDMDecoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)

    alignment = VFFoundationAlignment(latent_channels=z_channels, foundation_type="dinov2")
    return AutoencoderKL(encoder=encoder, decoder=decoder, alignment=alignment, **kwargs)


@register_model(f"continuous.vavae.f16_d16_mae",
                code_url="https://github.com/hustvl/LightningDiT/tree/2725fed42a14898744433809949834e26957bcdd",
                paper_url="https://arxiv.org/pdf/2501.01423",)
def VAVAE_f16_d16_mae(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=16,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 1, 2, 2, 4],
    num_res_blocks=2,
    attn_resolutions=[16],
    dropout=0.0,
    **kwargs):
    """
    VAVAE with compression factor 16
    Args:
        dims (int): Number of dimensions (2 for 2D, 3 for 3D)
    """
    encoder = LDMEncoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    decoder = LDMDecoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)

    alignment = VFFoundationAlignment(latent_channels=z_channels, foundation_type="mae")
    return AutoencoderKL(encoder=encoder, decoder=decoder, alignment=alignment, **kwargs)

@register_model(f"continuous.vavae.f16_d32_mae",
                code_url="https://github.com/hustvl/LightningDiT/tree/2725fed42a14898744433809949834e26957bcdd",
                paper_url="https://arxiv.org/pdf/2501.01423",)
def VAVAE_f16_d32_mae(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=32,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 1, 2, 2, 4],
    num_res_blocks=2,
    attn_resolutions=[16],
    dropout=0.0,
    **kwargs):
    return VAVAE_f16_d16_mae(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout, **kwargs)

@register_model(f"continuous.vavae.f16_d64_mae",
                code_url="https://github.com/hustvl/LightningDiT/tree/2725fed42a14898744433809949834e26957bcdd",
                paper_url="https://arxiv.org/pdf/2501.01423",)
def VAVAE_f16_d64_mae(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=64,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 1, 2, 2, 4],
    num_res_blocks=2,
    attn_resolutions=[16],
    dropout=0.0,
    **kwargs):
    return VAVAE_f16_d16_mae(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout, **kwargs)

@register_model(f"continuous.vavae.f16_d16_dinov2",
                code_url="https://github.com/hustvl/LightningDiT/tree/2725fed42a14898744433809949834e26957bcdd",
                paper_url="https://arxiv.org/pdf/2501.01423",)
def VAVAE_f16_d16_dinov2(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=16,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 1, 2, 2, 4],
    num_res_blocks=2,
    attn_resolutions=[16],
    dropout=0.0,
    **kwargs):
    """
    VAVAE with compression factor 16
    Args:
        dims (int): Number of dimensions (2 for 2D, 3 for 3D)
    """
    encoder = LDMEncoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)
    decoder = LDMDecoder(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout)

    alignment = VFFoundationAlignment(latent_channels=z_channels, foundation_type="dinov2")
    return AutoencoderKL(encoder=encoder, decoder=decoder, alignment=alignment, **kwargs)

@register_model(f"continuous.vavae.f16_d32_dinov2",
                code_url="https://github.com/hustvl/LightningDiT/tree/2725fed42a14898744433809949834e26957bcdd",
                paper_url="https://arxiv.org/pdf/2501.01423",)
def VAVAE_f16_d32_dinov2(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=32,
    in_channels=3,
    out_ch=3,
    ch=128,
    ch_mult=[1, 1, 2, 2, 4],
    num_res_blocks=2,
    attn_resolutions=[16],
    dropout=0.0,
    **kwargs):
    return VAVAE_f16_d16_dinov2(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout, **kwargs)

@register_model(f"continuous.vavae.f16_d64_dinov2",
                code_url="https://github.com/hustvl/LightningDiT/tree/2725fed42a14898744433809949834e26957bcdd",
                paper_url="https://arxiv.org/pdf/2501.01423",)
def VAVAE_f16_d64_dinov2(
    img_size=256,
    dims=2,
    ## Encoder decoder config
    double_z=True,
    z_channels=64,
    in_channels=3,
    out_ch=3,
    ch=128,
    
    ch_mult=[1, 1, 2, 2, 4],
    num_res_blocks=2,
    attn_resolutions=[16],
    dropout=0.0,
    **kwargs):
    return VAVAE_f16_d16_dinov2(img_size=img_size, dims=dims, double_z=double_z, z_channels=z_channels, in_channels=in_channels, out_ch=out_ch, ch=ch, ch_mult=ch_mult, num_res_blocks=num_res_blocks, attn_resolutions=attn_resolutions, dropout=dropout, **kwargs)


@register_model(
    f"continuous.dcae.f32c32",
    code_url="https://github.com/mit-han-lab/efficientvit",
    paper_url="https://arxiv.org/pdf/2410.10733",
    ckpt_path="https://huggingface.co/mit-han-lab/dc-ae-f32c32-in-1.0/resolve/main/model.safetensors",
)
def DCAE_f32c32(
    img_size=256,
    dims=2,
    in_channels=3,
    z_channels=32,
    width_list=(128, 256, 512, 512, 1024, 1024),
    depth_list=(0, 4, 8, 2, 2, 2),
    block_type=["ResBlock", "ResBlock", "ResBlock", "EViT_GLU", "EViT_GLU", "EViT_GLU"],
    norm="trms2d",
    act="silu",
    decoder_norm=["bn2d", "bn2d", "bn2d", "trms2d", "trms2d", "trms2d"],
    decoder_act=["relu", "relu", "relu", "silu", "silu", "silu"],
    decoder_depth_list=(0, 5, 10, 2, 2, 2),
    double_z=False,
    pre_post_layer="none",
    project_out_conv_only=True,
    **kwargs):
    """
    DCAE with compression factor 32 and 32 latent channels.
    double_z=False, use_quant_conv=False and project_out_conv_only=True match pretrained DCAE checkpoint keys.
    Args:
        dims (int): Number of dimensions (2 for 2D, 3 for 3D)
    """
    encoder = DCAEEncoder(
        in_channels=in_channels,
        z_channels=z_channels,
        width_list=width_list,
        depth_list=depth_list,
        block_type=block_type,
        norm=norm,
        act=act,
        double_z=double_z,
        dims=dims,
        img_size=img_size,
        project_out_conv_only=project_out_conv_only,
    )
    decoder = DCAEDecoder(
        in_channels=in_channels,
        z_channels=z_channels,
        width_list=width_list,
        depth_list=decoder_depth_list,
        block_type=block_type,
        norm=decoder_norm,
        act=decoder_act,
        dims=dims,
        img_size=img_size,
    )
    return AutoencoderKL(
        encoder=encoder,
        decoder=decoder,
        use_quant_conv=use_quant_conv,
        double_z=double_z,
        **kwargs,
    )


@register_model(
    f"continuous.dcae.f64c128",
    code_url="https://github.com/mit-han-lab/efficientvit",
    paper_url="https://arxiv.org/pdf/2410.10733",
    ckpt_path="https://huggingface.co/mit-han-lab/dc-ae-f64c128-in-1.0/resolve/main/model.safetensors",
)
def DCAE_f64c128(
    img_size=256,
    dims=2,
    in_channels=3,
    z_channels=128,
    width_list=(128, 256, 512, 512, 1024, 1024, 2048),
    depth_list=(0, 4, 8, 2, 2, 2, 2),
    block_type=["ResBlock", "ResBlock", "ResBlock", "EViT_GLU", "EViT_GLU", "EViT_GLU", "EViT_GLU"],
    norm="trms2d",
    act="silu",
    decoder_norm=["bn2d", "bn2d", "bn2d", "trms2d", "trms2d", "trms2d", "trms2d"],
    decoder_act=["relu", "relu", "relu", "silu", "silu", "silu", "silu"],
    decoder_depth_list=(0, 5, 10, 2, 2, 2, 2),
    double_z=False,
    pre_post_layer="none",
    project_out_conv_only=True,
    **kwargs):
    """
    DCAE with compression factor 64 and 128 latent channels
    """
    encoder = DCAEEncoder(
        in_channels=in_channels,
        z_channels=z_channels,
        width_list=width_list,
        depth_list=depth_list,
        block_type=block_type,
        norm=norm,
        act=act,
        double_z=double_z,
        dims=dims,
        img_size=img_size,
        project_out_conv_only=project_out_conv_only,
    )
    decoder = DCAEDecoder(
        in_channels=in_channels,
        z_channels=z_channels,
        width_list=width_list,
        depth_list=decoder_depth_list,
        block_type=block_type,
        norm=decoder_norm,
        act=decoder_act,
        dims=dims,
        img_size=img_size,
    )
    return AutoencoderKL(
        encoder=encoder, 
        decoder=decoder,
        double_z=double_z,
        use_quant_conv=use_quant_conv,
        **kwargs,
    )


@register_model(
    f"continuous.dcae.f128c512",
    code_url="https://github.com/mit-han-lab/efficientvit",
    paper_url="https://arxiv.org/pdf/2410.10733",
    ckpt_path="https://huggingface.co/mit-han-lab/dc-ae-f128c512-in-1.0/resolve/main/model.safetensors",
)
def DCAE_f128c512(
    img_size=256,
    dims=2,
    in_channels=3,
    z_channels=512,
    width_list=(128, 256, 512, 512, 1024, 1024, 2048, 2048),
    depth_list=(0, 4, 8, 2, 2, 2, 2, 2),
    block_type=["ResBlock", "ResBlock", "ResBlock", "EViT_GLU", "EViT_GLU", "EViT_GLU", "EViT_GLU", "EViT_GLU"],
    norm="trms2d",
    act="silu",
    decoder_norm=["bn2d", "bn2d", "bn2d", "trms2d", "trms2d", "trms2d", "trms2d", "trms2d"],
    decoder_act=["relu", "relu", "relu", "silu", "silu", "silu", "silu", "silu"],
    decoder_depth_list=(0, 5, 10, 2, 2, 2, 2, 2),
    double_z=False,
    pre_post_layer="none",
    project_out_conv_only=True,
    **kwargs):
    """
    DCAE with compression factor 128 and 512 latent channels
    """
    encoder = DCAEEncoder(
        in_channels=in_channels,
        z_channels=z_channels,
        width_list=width_list,
        depth_list=depth_list,
        block_type=block_type,
        norm=norm,
        act=act,
        double_z=double_z,
        dims=dims,
        img_size=img_size,
        project_out_conv_only=project_out_conv_only,
    )
    decoder = DCAEDecoder(
        in_channels=in_channels,
        z_channels=z_channels,
        width_list=width_list,
        depth_list=decoder_depth_list,
        block_type=block_type,
        norm=decoder_norm,
        act=decoder_act,
        dims=dims,
        img_size=img_size,
        )
    return AutoencoderKL(
        encoder=encoder,
        decoder=decoder,
        double_z=double_z,
        use_quant_conv=use_quant_conv,
        **kwargs,
    )


@register_model(f"discrete.hcvq.vae.S_16")
def SDVQ_Transformer_S_16(
    img_size: int = 256,
    patch_size: int = 16,
    in_channels: int = 3,
    embed_dim_encoder: int = 384,
    embed_dim_decoder: int = 384,
    depth_encoder: int = 12,
    depth_decoder: int = 12,
    num_heads_encoder: int = 6,
    num_heads_decoder: int = 6,
    mlp_ratio: float = 4.0,
    mask_ratio_mu: float = 0.0,
    masking: str = "none",
    e_dim: int = 32,
    **kwargs
    ):

    encoder = GenericViTEncoder(
        img_size=img_size,
        patch_size=patch_size,
        in_channels=in_channels,
        embed_dim=embed_dim_encoder,
        depth=depth_encoder,
        num_heads=num_heads_encoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        mask_ratio_mu=mask_ratio_mu,
        masking=masking,
        double_z=True,
    )
    decoder = GenericViTDecoder(
        img_size=img_size,
        patch_size=patch_size,
        out_channels=in_channels,
        embed_dim=embed_dim_decoder,
        depth=depth_decoder,
        num_heads=num_heads_decoder,
        mlp_ratio=mlp_ratio,
        pos_type="learned",
        use_rope=True,
        num_prefix_tokens=1,
        num_latent_tokens=0,
        to_pixel="conv",
        token_dim=None,   # PostQuantLayer Is Done in the VQModel
    )

    return AutoencoderKLTransformer(encoder=encoder, decoder=decoder, pre_post_layer="linear", embed_dim=e_dim, channel_dim=2, **kwargs)


from __future__ import annotations

import importlib
from typing import Any, Mapping, Literal

import torch
import torch.nn as nn


def get_model_type(model: nn.Module) -> Literal["continuous", "discrete", "token", "autoregressive", "non-autoregressive"]:
    module_path = model.__class__.__module__
    if "first_stage.continuous" in module_path:
        return "continuous"
    elif "first_stage.discrete" in module_path:
        return "discrete"
    elif "first_stage.token" in module_path:
        return "token"
    elif "generators.autoregressive" in module_path:
        return "autoregressive"
    elif "generators.non_autoregressive" in module_path:
        return "non-autoregressive"
    else:
        raise ValueError(f"Unknown model type: {module_path}")

from typing import Any
import hashlib
import os
import urllib.request
from urllib.parse import urlparse

import torch


def _resolve_ckpt_path(path: str) -> str:
    """
    Resolve ``path`` to a local file path. If ``path`` is an http(s) URL, the
    file is downloaded to a cache directory and the local path is returned.
    Works with any direct download link (Hugging Face, Google Drive, etc.).
    """
    if os.path.isfile(path):
        return path
    if not path.startswith(("http://", "https://")):
        return path

    # Download from URL to cache
    parsed = urlparse(path)
    ext = os.path.splitext(parsed.path)[1] or ".bin"
    cache_key = hashlib.sha256(path.encode()).hexdigest()
    cache_dir = os.path.join(os.path.expanduser("~"), ".cache", "medlat", "downloads")
    os.makedirs(cache_dir, exist_ok=True)
    cached_path = os.path.join(cache_dir, cache_key + ext)

    if os.path.isfile(cached_path):
        return cached_path

    try:
        request = urllib.request.Request(path, headers={"User-Agent": "medlat/1.0"})
        with urllib.request.urlopen(request) as resp:
            with open(cached_path, "wb") as f:
                while True:
                    chunk = resp.read(8192)
                    if not chunk:
                        break
                    f.write(chunk)
    except Exception as e:
        if os.path.isfile(cached_path):
            try:
                os.remove(cached_path)
            except OSError:
                pass
        raise RuntimeError(f"Failed to download checkpoint from {path!r}: {e}") from e

    if not os.path.isfile(cached_path):
        raise FileNotFoundError(f"Download completed but file missing: {cached_path!r}")
    return cached_path


def init_from_ckpt(model, path: str, weights_only: bool = False) -> None:
    """
    Load a checkpoint into ``model`` while remaining tolerant to shape mismatches.
    Supports .ckpt/.pt and .safetensors files. If ``path`` is an http(s) URL, the
    checkpoint is downloaded and cached under ~/.cache/medlat/downloads.
    """
    path = _resolve_ckpt_path(path)
    if not os.path.isfile(path):
        raise FileNotFoundError(
            f"Checkpoint path is not an existing file: {path!r}. "
            "Use a local path or an http(s) download URL."
        )

    is_safetensors = "safetensor" in path

    def _load_torch() -> dict[str, Any]:
        return torch.load(path, map_location="cpu", weights_only=weights_only)

    def _load_safetensors() -> dict[str, Any]:
        from safetensors.torch import load_file
        return load_file(path, device="cpu")

    def _load_candidate(candidate_key: str) -> dict[str, Any] | None:
        if is_safetensors:
            # safetensors are always flat state_dicts
            return None
        try:
            return _load_torch()[candidate_key]
        except Exception:
            return None

    # --- Load state_dict ---
    if is_safetensors:
        state_dict = _load_safetensors()
    else:
        state_dict = (
            _load_candidate("state_dict")
            or _load_candidate("model")
            or _load_torch()
        )

    # --- Clean keys ---
    cleaned = {}
    for key, value in state_dict.items():
        if "loss" in key:
            continue
        new_key = key.replace("module.", "", 1) if key.startswith("module.") else key
        cleaned[new_key] = value

    # --- Load into model ---
    try:
        msg = model.load_state_dict(cleaned, strict=True)
    except RuntimeError as err:
        print(f"Warning: {err}")
        print("Falling back to non-strict loading (often happens when mixing 2D/3D weights).")
        msg = model.load_state_dict(cleaned, strict=False)

    torch.cuda.empty_cache()

    print(f"Loading pre-trained {model.__class__.__name__}")
    print("Missing keys:")
    print(msg.missing_keys)
    print("Unexpected keys:")
    print(msg.unexpected_keys)
    print(f"Restored from {path}")


def instantiate_from_config(config: Mapping[str, Any]) -> Any:
    """
    Instantiate an object from a Hydra/OmegaConf-style configuration dict.

    The dictionary must contain a ``_target_`` entry with the fully qualified
    import path to the callable. Any additional key/value pairs are forwarded
    as keyword arguments.
    """

    if "_target_" not in config:
        raise KeyError("Configuration dictionary must contain a '_target_' entry.")

    target = config["_target_"]
    module_name, attr_name = target.rsplit(".", 1)
    module = importlib.import_module(module_name)
    callable_obj = getattr(module, attr_name)
    kwargs = {k: v for k, v in config.items() if k != "_target_"}
    return callable_obj(**kwargs)



from .backbone import *
from .cls import *
from .dc_ae import *
# Import sam and seg only if dependencies are available
try:
    from .sam import *
except ImportError:
    pass
try:
    from .seg import *
except ImportError:
    pass



from typing import Optional

import torch
import torch.nn as nn

from efficientvit.models.efficientvit.backbone import EfficientViTBackbone, EfficientViTLargeBackbone
from efficientvit.models.nn import (
    ConvLayer,
    DAGBlock,
    FusedMBConv,
    IdentityLayer,
    MBConv,
    OpSequential,
    ResidualBlock,
    UpSampleLayer,
)
from efficientvit.models.utils import build_kwargs_from_config

__all__ = [
    "EfficientViTSeg",
    "efficientvit_seg_b0",
    "efficientvit_seg_b1",
    "efficientvit_seg_b2",
    "efficientvit_seg_b3",
    "efficientvit_seg_l1",
    "efficientvit_seg_l2",
]


class SegHead(DAGBlock):
    def __init__(
        self,
        fid_list: list[str],
        in_channel_list: list[int],
        stride_list: list[int],
        head_stride: int,
        head_width: int,
        head_depth: int,
        expand_ratio: float,
        middle_op: str,
        final_expand: Optional[float],
        n_classes: int,
        dropout=0,
        norm="bn2d",
        act_func="hswish",
    ):
        inputs = {}
        for fid, in_channel, stride in zip(fid_list, in_channel_list, stride_list):
            factor = stride // head_stride
            if factor == 1:
                inputs[fid] = ConvLayer(in_channel, head_width, 1, norm=norm, act_func=None)
            else:
                inputs[fid] = OpSequential(
                    [
                        ConvLayer(in_channel, head_width, 1, norm=norm, act_func=None),
                        UpSampleLayer(factor=factor),
                    ]
                )

        middle = []
        for _ in range(head_depth):
            if middle_op == "mbconv":
                block = MBConv(
                    head_width,
                    head_width,
                    expand_ratio=expand_ratio,
                    norm=norm,
                    act_func=(act_func, act_func, None),
                )
            elif middle_op == "fmbconv":
                block = FusedMBConv(
                    head_width,
                    head_width,
                    expand_ratio=expand_ratio,
                    norm=norm,
                    act_func=(act_func, None),
                )
            else:
                raise NotImplementedError
            middle.append(ResidualBlock(block, IdentityLayer()))
        middle = OpSequential(middle)

        outputs = {
            "segout": OpSequential(
                [
                    (
                        None
                        if final_expand is None
                        else ConvLayer(head_width, head_width * final_expand, 1, norm=norm, act_func=act_func)
                    ),
                    ConvLayer(
                        head_width * (1 if final_expand is None else final_expand),
                        n_classes,
                        1,
                        use_bias=True,
                        dropout=dropout,
                        norm=None,
                        act_func=None,
                    ),
                ]
            )
        }

        super(SegHead, self).__init__(inputs, "add", None, middle=middle, outputs=outputs)


class EfficientViTSeg(nn.Module):
    def __init__(self, backbone: EfficientViTBackbone | EfficientViTLargeBackbone, head: SegHead) -> None:
        super().__init__()
        self.backbone = backbone
        self.head = head

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        feed_dict = self.backbone(x)
        feed_dict = self.head(feed_dict)

        return feed_dict["segout"]


def efficientvit_seg_b0(dataset: str, **kwargs) -> EfficientViTSeg:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_b0

    backbone = efficientvit_backbone_b0(**kwargs)

    if dataset == "cityscapes":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[128, 64, 32],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=32,
            head_depth=1,
            expand_ratio=4,
            middle_op="mbconv",
            final_expand=4,
            n_classes=19,
            **build_kwargs_from_config(kwargs, SegHead),
        )
    else:
        raise NotImplementedError
    model = EfficientViTSeg(backbone, head)
    return model


def efficientvit_seg_b1(dataset: str, **kwargs) -> EfficientViTSeg:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_b1

    backbone = efficientvit_backbone_b1(**kwargs)

    if dataset == "cityscapes":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[256, 128, 64],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=64,
            head_depth=3,
            expand_ratio=4,
            middle_op="mbconv",
            final_expand=4,
            n_classes=19,
            **build_kwargs_from_config(kwargs, SegHead),
        )
    elif dataset == "ade20k":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[256, 128, 64],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=64,
            head_depth=3,
            expand_ratio=4,
            middle_op="mbconv",
            final_expand=None,
            n_classes=150,
            **build_kwargs_from_config(kwargs, SegHead),
        )
    else:
        raise NotImplementedError
    model = EfficientViTSeg(backbone, head)
    return model


def efficientvit_seg_b2(dataset: str, **kwargs) -> EfficientViTSeg:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_b2

    backbone = efficientvit_backbone_b2(**kwargs)

    if dataset == "cityscapes":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[384, 192, 96],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=96,
            head_depth=3,
            expand_ratio=4,
            middle_op="mbconv",
            final_expand=4,
            n_classes=19,
            **build_kwargs_from_config(kwargs, SegHead),
        )
    elif dataset == "ade20k":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[384, 192, 96],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=96,
            head_depth=3,
            expand_ratio=4,
            middle_op="mbconv",
            final_expand=None,
            n_classes=150,
            **build_kwargs_from_config(kwargs, SegHead),
        )
    else:
        raise NotImplementedError
    model = EfficientViTSeg(backbone, head)
    return model


def efficientvit_seg_b3(dataset: str, **kwargs) -> EfficientViTSeg:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_b3

    backbone = efficientvit_backbone_b3(**kwargs)

    if dataset == "cityscapes":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[512, 256, 128],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=128,
            head_depth=3,
            expand_ratio=4,
            middle_op="mbconv",
            final_expand=4,
            n_classes=19,
            **build_kwargs_from_config(kwargs, SegHead),
        )
    elif dataset == "ade20k":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[512, 256, 128],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=128,
            head_depth=3,
            expand_ratio=4,
            middle_op="mbconv",
            final_expand=None,
            n_classes=150,
            **build_kwargs_from_config(kwargs, SegHead),
        )
    else:
        raise NotImplementedError
    model = EfficientViTSeg(backbone, head)
    return model


def efficientvit_seg_l1(dataset: str, **kwargs) -> EfficientViTSeg:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_l1

    backbone = efficientvit_backbone_l1(**kwargs)

    if dataset == "cityscapes":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[512, 256, 128],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=256,
            head_depth=3,
            expand_ratio=1,
            middle_op="fmbconv",
            final_expand=None,
            n_classes=19,
            act_func="gelu",
            **build_kwargs_from_config(kwargs, SegHead),
        )
    elif dataset == "ade20k":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[512, 256, 128],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=128,
            head_depth=3,
            expand_ratio=4,
            middle_op="fmbconv",
            final_expand=8,
            n_classes=150,
            act_func="gelu",
            **build_kwargs_from_config(kwargs, SegHead),
        )
    else:
        raise NotImplementedError
    model = EfficientViTSeg(backbone, head)
    return model


def efficientvit_seg_l2(dataset: str, **kwargs) -> EfficientViTSeg:
    from efficientvit.models.efficientvit.backbone import efficientvit_backbone_l2

    backbone = efficientvit_backbone_l2(**kwargs)

    if dataset == "cityscapes":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[512, 256, 128],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=256,
            head_depth=5,
            expand_ratio=1,
            middle_op="fmbconv",
            final_expand=None,
            n_classes=19,
            act_func="gelu",
            **build_kwargs_from_config(kwargs, SegHead),
        )
    elif dataset == "ade20k":
        head = SegHead(
            fid_list=["stage4", "stage3", "stage2"],
            in_channel_list=[512, 256, 128],
            stride_list=[32, 16, 8],
            head_stride=8,
            head_width=128,
            head_depth=3,
            expand_ratio=4,
            middle_op="fmbconv",
            final_expand=8,
            n_classes=150,
            act_func="gelu",
            **build_kwargs_from_config(kwargs, SegHead),
        )
    else:
        raise NotImplementedError
    model = EfficientViTSeg(backbone, head)
    return model



from __future__ import annotations
import torch
from typing import Tuple, Sequence
from einops import rearrange
from einops.layers.torch import Rearrange
import torch.nn as nn
import numpy as np


def _to_tuple(value: int | Sequence[int], dims: int) -> Tuple[int, ...]:
    """
    Normalize an integer / sequence into a tuple of length ``dims``.

    Accepts:
    - scalar int: replicated ``dims`` times
    - tuple / list of ints: must already have length ``dims``
    """
    if isinstance(value, (tuple, list)):
        if len(value) != dims:
            raise ValueError(f"Expected {dims} values, got {len(value)}")
        return tuple(int(v) for v in value)
    return (int(value),) * dims


def _infer_dims(img_size: int | Tuple[int, ...]) -> int:
    return 3 if isinstance(img_size, tuple) and len(img_size) == 3 else 2


class SineLayer(nn.Module):
    """
    Paper: Implicit Neural Representation with Periodic Activ ation Function (SIREN)
    """

    def __init__(self, in_features, out_features, bias=True, is_first=False, omega_0=30):
        super().__init__()
        self.omega_0 = omega_0
        self.is_first = is_first

        self.in_features = in_features
        self.linear = nn.Linear(in_features, out_features, bias=bias)

        self.init_weights()

    def init_weights(self):
        with torch.no_grad():
            if self.is_first:
                self.linear.weight.uniform_(-1 / self.in_features, 1 / self.in_features)
            else:
                self.linear.weight.uniform_(-np.sqrt(6 / self.in_features) / self.omega_0,
                                            np.sqrt(6 / self.in_features) / self.omega_0)

    def forward(self, input):
        return torch.sin(self.omega_0 * self.linear(input))


class PatchEmbed(nn.Module):
    """Image/Volume to Patch Embedding - Supports both 2D and 3D"""
    def __init__(self, to_embed='linear', img_size=224, patch_size=16, in_chans=3, embed_dim=768, norm_layer=None, flatten=True):
        super().__init__()
        
        # Handle img_size: can be int (2D), tuple of 2 (2D), or tuple of 3 (3D)
        if isinstance(img_size, int):
            self.img_size = (img_size, img_size)
            self.dims = 2
        elif len(img_size) == 2:
            self.img_size = img_size
            self.dims = 2
        elif len(img_size) == 3:
            self.img_size = img_size
            self.dims = 3
        else:
            raise ValueError(f"img_size must be int or tuple of length 2 or 3, got {img_size}")
        
        # Handle patch_size: can be int (2D), tuple of 2 (2D), or tuple of 3 (3D)
        if isinstance(patch_size, int):
            if self.dims == 2:
                self.patch_size = (patch_size, patch_size)
            else:  # 3D
                self.patch_size = (patch_size, patch_size, patch_size)
        elif len(patch_size) == 2:
            if self.dims == 2:
                self.patch_size = patch_size
            else:
                raise ValueError(f"patch_size tuple of length 2 not compatible with 3D img_size {img_size}")
        elif len(patch_size) == 3:
            if self.dims == 3:
                self.patch_size = patch_size
            else:
                raise ValueError(f"patch_size tuple of length 3 not compatible with 2D img_size {img_size}")
        else:
            raise ValueError(f"patch_size must be int or tuple of length 2 or 3, got {patch_size}")
        
        # Calculate grid size and number of patches
        if self.dims == 2:
            self.grid_size = (self.img_size[0] // self.patch_size[0], self.img_size[1] // self.patch_size[1])
            self.num_patches = self.grid_size[0] * self.grid_size[1]
        else:  # 3D
            self.grid_size = (self.img_size[0] // self.patch_size[0], 
                             self.img_size[1] // self.patch_size[1], 
                             self.img_size[2] // self.patch_size[2])
            self.num_patches = self.grid_size[0] * self.grid_size[1] * self.grid_size[2]
        
        self.flatten = flatten
        
        if to_embed == 'conv':
            # Create appropriate convolution layer
            if self.dims == 2:
                self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=self.patch_size, stride=self.patch_size)
            else:  # 3D
                self.proj = nn.Conv3d(in_chans, embed_dim, kernel_size=self.patch_size, stride=self.patch_size)
        
        elif to_embed == 'siren':
            self.proj = nn.Sequential(
                    SineLayer(in_chans, embed_dim, is_first=True, omega_0=30.),
                    SineLayer(embed_dim, embed_dim * 2, is_first=False, omega_0=30)
                )
        
        elif to_embed == 'linear':
            self.proj = nn.Linear(in_chans, embed_dim)
        elif to_embed == 'identity':
            self.proj = nn.Identity()
        else:
            raise NotImplementedError

        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()


    def forward(self, x):
        if self.dims == 2:
            B, C, H, W = x.shape
            assert H == self.img_size[0] and W == self.img_size[1], \
                f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]})."
        else:  # 3D
            B, C, D, H, W = x.shape
            assert D == self.img_size[0] and H == self.img_size[1] and W == self.img_size[2], \
                f"Input volume size ({D}*{H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]}*{self.img_size[2]})."
        
        x = self.proj(x)
        
        if self.flatten:
            if self.dims == 2:
                x = x.flatten(2).transpose(1, 2)  # BCHW -> BNC
            else:  # 3D
                x = x.flatten(2).transpose(1, 2)  # BCDHW -> BNC
        
        x = self.norm(x)
        return x


class ToPixel(nn.Module):
    def __init__(self, to_pixel='linear', img_size=256, out_channels=3, in_dim=512, patch_size=16) -> None:
        super().__init__()
        self.to_pixel_name = to_pixel
        
        # Handle img_size: can be int (2D), tuple of 2 (2D), or tuple of 3 (3D)
        if isinstance(img_size, int):
            self.img_size = (img_size, img_size)
            self.dims = 2
        elif len(img_size) == 2:
            self.img_size = img_size
            self.dims = 2
        elif len(img_size) == 3:
            self.img_size = img_size
            self.dims = 3
        else:
            raise ValueError(f"img_size must be int or tuple of length 2 or 3, got {img_size}")
        
        # Handle patch_size: can be int (2D), tuple of 2 (2D), or tuple of 3 (3D)
        if isinstance(patch_size, int):
            if self.dims == 2:
                self.patch_size = (patch_size, patch_size)
            else:  # 3D
                self.patch_size = (patch_size, patch_size, patch_size)
        elif len(patch_size) == 2:
            if self.dims == 2:
                self.patch_size = patch_size
            else:
                raise ValueError(f"patch_size tuple of length 2 not compatible with 3D img_size {img_size}")
        elif len(patch_size) == 3:
            if self.dims == 3:
                self.patch_size = patch_size
            else:
                raise ValueError(f"patch_size tuple of length 3 not compatible with 2D img_size {img_size}")
        else:
            raise ValueError(f"patch_size must be int or tuple of length 2 or 3, got {patch_size}")
        
        # Calculate grid size and number of patches
        if self.dims == 2:
            self.grid_size = (self.img_size[0] // self.patch_size[0], self.img_size[1] // self.patch_size[1])
            self.num_patches = self.grid_size[0] * self.grid_size[1]
        else:  # 3D
            self.grid_size = (self.img_size[0] // self.patch_size[0], 
                             self.img_size[1] // self.patch_size[1], 
                             self.img_size[2] // self.patch_size[2])
            self.num_patches = self.grid_size[0] * self.grid_size[1] * self.grid_size[2]

        self.out_channels = out_channels
        if to_pixel == 'linear':
            if self.dims == 2:
                self.model = nn.Linear(in_dim, out_channels * self.patch_size[0] * self.patch_size[1])
            else:  # 3D
                self.model = nn.Linear(in_dim, out_channels * self.patch_size[0] * self.patch_size[1] * self.patch_size[2])
        elif to_pixel == 'conv':
            if self.dims == 2:
                self.model = nn.Sequential(
                    Rearrange("b (h w) c -> b c h w", h=self.grid_size[0], w=self.grid_size[1]),
                    nn.Conv2d(in_dim, self.patch_size[0] * self.patch_size[1] * out_channels, 1, padding=0),
                    Rearrange("b (p1 p2 c) h w -> b c (h p1) (w p2)", p1=self.patch_size[0], p2=self.patch_size[1]),
                    nn.Conv2d(out_channels, out_channels, 3, padding=1)
                    )
            else:  # 3D
                self.model = nn.Sequential(
                    Rearrange("b (d h w) c -> b c d h w", d=self.grid_size[0], h=self.grid_size[1], w=self.grid_size[2]),
                    nn.Conv3d(in_dim, self.patch_size[0] * self.patch_size[1] * self.patch_size[2] * out_channels, 1, padding=0),
                    Rearrange("b (p1 p2 p3 c) d h w -> b c (d p1) (h p2) (w p3)", p1=self.patch_size[0], p2=self.patch_size[1], p3=self.patch_size[2]),
                    nn.Conv3d(out_channels, out_channels, 3, padding=1)
                    )
        elif to_pixel == 'siren':
            if self.dims == 2:
                self.model = nn.Sequential(
                    SineLayer(in_dim, in_dim * 2, is_first=True, omega_0=30.),
                    SineLayer(in_dim * 2, self.img_size[0] // self.patch_size[0] * self.patch_size[0] * out_channels, is_first=False, omega_0=30)
                )
            else:  # 3D
                self.model = nn.Sequential(
                    SineLayer(in_dim, in_dim * 2, is_first=True, omega_0=30.),
                    SineLayer(in_dim * 2, self.img_size[0] // self.patch_size[0] * self.patch_size[0] * out_channels, is_first=False, omega_0=30)
                )
        elif to_pixel == 'identity' or to_pixel == 'none':
            self.model = nn.Identity()
        else:
            raise NotImplementedError

    def get_last_layer(self):
        if self.to_pixel_name == 'linear':
            return self.model.weight
        elif self.to_pixel_name == 'siren':
            return self.model[1].linear.weight
        elif self.to_pixel_name == 'conv':
            return self.model[-1].weight
        else:
            return None

    def unpatchify(self, x):
        """
        x: (N, L, patch_size**2 * out_channels) for 2D 
        (N, L, patch_size**3 * out_channels) for 3D
        imgs: (N, out_channels, H, W) for 2D 
            (N, out_channels, D, H, W) for 3D
        """
        B = x.shape[0]
        C = self.out_channels

        if self.dims == 2:
            p_h, p_w = self.patch_size
            H, W = self.img_size
            h, w = H // p_h, W // p_w
            assert h * w == x.shape[1], f"Expected {h*w} patches, got {x.shape[1]}"
            imgs = rearrange(
                x, "b (h w) (ph pw c) -> b c (h ph) (w pw)",
                h=h, w=w, ph=p_h, pw=p_w, c=C
            )

        else:  # 3D
            p_d, p_h, p_w = self.patch_size
            D, H, W = self.img_size
            d, h, w = D // p_d, H // p_h, W // p_w
            assert d * h * w == x.shape[1], f"Expected {d*h*w} patches, got {x.shape[1]}"
            imgs = rearrange(
                x, "b (d h w) (pd ph pw c) -> b c (d pd) (h ph) (w pw)",
                d=d, h=h, w=w, pd=p_d, ph=p_h, pw=p_w, c=C
            )
        return imgs


    def forward(self, x):
        if self.to_pixel_name == 'linear':
            x = self.model(x)
            x = self.unpatchify(x)
        elif self.to_pixel_name == 'siren':
            x = self.model(x)
            if self.dims == 2:
                x = x.view(x.shape[0], self.out_channels, self.patch_size[0] * int(self.num_patches ** 0.5),
                           self.patch_size[0] * int(self.num_patches ** 0.5))
            else:  # 3D
                d = self.img_size[0] // self.patch_size[0]
                h = self.img_size[1] // self.patch_size[1]
                w = self.img_size[2] // self.patch_size[2]
                x = x.view(x.shape[0], self.out_channels, self.patch_size[0] * d,
                           self.patch_size[1] * h, self.patch_size[2] * w)
        elif self.to_pixel_name == 'conv':
            x = self.model(x)
        elif self.to_pixel_name == 'identity':
            x = self.model(x)
        elif self.to_pixel_name == 'none':
            x = self.model(x)
            x = self.unpatchify(x)
        return x